/*
 * Copyright (c) 2004-2005 Northwestern University.
 *                         NuCAD Group.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * acg_base.h - Data structure for bi-colored directed graphs.
 *
 * Authors: Hai Zhou, haizhou@ece.northwestern.edu
 *          Jia Wang, jwa112@ece.northwestern.edu
 */

#ifndef ACG_BASE_H
#define ACG_BASE_H

#include "../misc/geometry.h"

namespace acg
{

// fanins/fanouts, from/to
enum pos_t {POS_IN = 0, POS_OUT = 1};
inline pos_t other(pos_t pos) {return pos_t(1-pos);}

// horizental/vertical
enum color_t {COLOR_RED = 0, COLOR_BLACK = 1};
inline color_t other(color_t c) {return color_t(1-c);}

// forwards
class node_t;

class edge_t
{
	friend class edge_list_t;
	friend class acg_dev_t;

	// no copy, no assign
	edge_t(const edge_t &);
	edge_t &operator = (const edge_t &);

	edge_t *next_[2], *prev_[2]; // 0: same fanin, 1: same fanout
	
	node_t *node_[2]; // 0: from, 1: to

	edge_t() : m_data(0)
	{
		node_[0] = node_[1] = 0;
		next_[0] = next_[1] = prev_[0] = prev_[1] = this;
	}

	~edge_t() {}

public:

	mutable void *m_data; // scratching data for edges

	edge_t *next(pos_t pos) {return next_[pos];}
	edge_t *prev(pos_t pos) {return prev_[pos];}
	node_t *node(pos_t pos) {return node_[pos];}

	const edge_t *next(pos_t pos) const {return next_[pos];}
	const edge_t *prev(pos_t pos) const {return prev_[pos];}
	const node_t *node(pos_t pos) const {return node_[pos];}

	bool valid() const;

}; // class edge_t

class edge_list_t
{
	friend class node_t;

	// no copy, no assign
	edge_list_t(const edge_list_t &);
	edge_list_t &operator = (const edge_list_t &);

	edge_t head_; // two double linked circular list, cross linked with other lists

	// need this to init array elements...
	void init(node_t *n)
	{
		head_.node_[0] = head_.node_[1] = n;
	}

public:
	edge_list_t() {}

	node_t *node() {return head_.node(POS_IN);}
	const node_t *node() const {return head_.node(POS_IN);}

	edge_t *begin(pos_t pos) {return head_.next(pos);}
	edge_t *end(pos_t pos) {return &head_;}
	const edge_t *begin(pos_t pos) const {return head_.next(pos);}
	const edge_t *end(pos_t pos) const {return &head_;}

	bool empty(pos_t pos) const {return begin(pos) == end(pos);}

	typedef std::pair<edge_t *, edge_t *> range_t;
	typedef std::pair<const edge_t *, const edge_t *> const_range_t;
	range_t all(pos_t pos) {return std::make_pair(begin(pos), end(pos));}
	const_range_t all(pos_t pos) const {return std::make_pair(begin(pos), end(pos));}

	edge_t *front(pos_t pos) {assert(!empty(pos)); return head_.next(pos);}
	edge_t *back(pos_t pos) {assert(!empty(pos)); return head_.prev(pos);}
	const edge_t *front(pos_t pos) const {assert(!empty(pos)); return head_.next(pos);}
	const edge_t *back(pos_t pos) const {assert(!empty(pos)); return head_.next(pos);}

	edge_t *pop_front(pos_t pos);
	edge_t *pop_back(pos_t pos);
	void push_front(edge_t *edge, pos_t pos);
	void push_back(edge_t *edge, pos_t pos);

	// The following functions are all static because we don't keep the size of the edge list.

	// Insert empty 'edge' before 'after' on 'pos' side; return 'edge'.
	static void insert(edge_t *after, edge_t *edge, pos_t pos);

	// Erase 'edge' from 'pos' side.
	static void erase(edge_t *edge, pos_t pos);

	// Splice ['first', 'last') before 'after' at 'pos' side.
	static void splice(edge_t *after, edge_t *first, edge_t *last, pos_t pos);

	bool valid() const;

}; // class edge_list_t

class node_t
{
	friend class acg_dev_t;

	// no copy, no assign
	node_t(const node_t &);
	node_t &operator = (const node_t &);

	edge_list_t edges_[2]; // COLOR_BLACK, COLOR_RED

	double width_[2];
	double bb_[2][2]; // color/pos

	node_t();
	node_t(double w, double h, size_t id);
	~node_t() {}

public:
	size_t m_id; // user managed ids

	mutable void *m_data; // scratching data for nodes
	// more scratching data from edge list
	void *&edge_data(color_t c) const {return edges_[c].head_.m_data;}

	edge_list_t *edges(color_t c) {return &edges_[c];}
	const edge_list_t *edges(color_t c) const {return &edges_[c];}

	// several shortcuts to get boundary node directly
	// return self if no such node exists
	
	node_t *nearest_node(color_t c, pos_t pos)
		{return edges(c)->begin(pos)->node(other(pos));}
	const node_t *nearest_node(color_t c, pos_t pos) const
		{return edges(c)->begin(pos)->node(other(pos));}

	node_t *farest_node(color_t c, pos_t pos)
		{return edges(c)->end(pos)->prev(pos)->node(other(pos));}
	const node_t *farest_node(color_t c, pos_t pos) const
		{return edges(c)->end(pos)->prev(pos)->node(other(pos));}

	bool valid() const;

	// shape and packing
	// use functions so that you don't need to know how to index

	double &width(color_t c) {return width_[c];}
	double width(color_t c) const {return width_[c];}

	double &bb(color_t c, pos_t pos) {return bb_[c][pos];}
	double bb(color_t c, pos_t pos) const {return bb_[c][pos];}

	double boundary(color_t c, pos_t pos) const {return bb(c, pos)+width(c);}

	// tuples
	misc::dbl_pair width() const
		{return misc::dbl_pair(width_[COLOR_RED], width_[COLOR_BLACK]);}
	misc::dbl_pair bb(pos_t pos) const
		{return misc::dbl_pair(bb_[COLOR_RED][pos], bb_[COLOR_BLACK][pos]);}
	misc::dbl_pair boundary(pos_t pos) const {return bb(pos)+width();}
	misc::dbl_pair center(pos_t pos) const {return bb(pos)+width()/2;}

	// copy/swap shape, bb, and id
	void copy_content(const node_t &src);
	void swap_content(node_t &src);

}; // class node_t

class acg_base_t
{
	friend class acg_dev_t;

	size_t n_edges_;
	std::vector<node_t *> nodes_;

public:
	// create an empty one
	acg_base_t();
	// initialized as empty for direct building
	acg_base_t(size_t n, misc::dbl_pair shapes[], size_t ids[] = 0);

	~acg_base_t();
	void clear();

	// swap, copy, and assign
	void swap(acg_base_t &r);
	// node_t::m_data and edge_t::m_data will be overwrited
	acg_base_t(const acg_base_t &r);
	acg_base_t &operator = (acg_base_t r) {swap(r); return *this;}

	// statistics
	size_t num_nodes() const {return nodes_.size();}
	size_t num_edges() const {return n_edges_;}
	bool empty() const {return nodes_.empty();}

	// individual node visiting
	node_t *at(size_t i) {assert(i < num_nodes()); return nodes_[i];}
	const node_t *at(size_t i) const {assert(i < num_nodes()); return nodes_[i];}
	node_t *operator[](size_t i) {return at(i);}
	const node_t *operator[](size_t i) const {return at(i);}

	// write modules and edges using pointer as names
	// type default to ACG-BASE
	void xml_write(FILE *fp, const char *type = "ACG-BASE");
	// However, edge read in progress...
	// bool? xml_read(FILE *fp, ???);

	// node perturbations
	void node_swap(size_t i, size_t j);
	void node_rotate(size_t where);
	void node_resize(size_t where, double w, double h);

}; // class acg_base_t

} // namespace acg

namespace std
{

template<> inline void swap(acg::acg_base_t &a, acg::acg_base_t &b) {a.swap(b);}

} // namespace std

#endif // ACG_BASE_H
