/*
 * Copyright (c) 2004-2005 Northwestern University.
 *                         NuCAD Group.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * acg_dev.cpp - Internal implementation details.
 *
 * Authors: Hai Zhou, haizhou@ece.northwestern.edu
 *          Jia Wang, jwa112@ece.northwestern.edu
 */

#include "../misc/common.h"
#include "../misc/malloc_pool.h"

// for placement new
#include <new>

#include "acg_dev.h"

namespace acg
{

namespace detail
{

typedef misc::malloc_pool<sizeof(edge_t)> edge_pool_t;
typedef misc::malloc_pool<sizeof(node_t)> node_pool_t;

// singletons for pools
edge_pool_t *get_edge_pool()
{
	static edge_pool_t pool;
	return &pool;
}

node_pool_t *get_node_pool()
{
	static node_pool_t pool;
	return &pool;
}

} // namespace detail

// The following 5 functions are where you should look for
// to add diagnostic code if you mess things up.

edge_t *acg_dev_t::alloc_edge()
{
	++n_edges_;

	// there will be a leakage for ctor exceptions
	// however, edge_t is so simple that it will never
	// generate an exception... which applies to node_t too.
	return new (detail::get_edge_pool()->alloc()) edge_t;
}

void acg_dev_t::free_edge(edge_t *edge)
{
	--n_edges_;

	edge->~edge_t();

	detail::get_edge_pool()->free(edge);
}

node_t *acg_dev_t::alloc_node()
{
	return new (detail::get_node_pool()->alloc()) node_t;
}

node_t *acg_dev_t::alloc_node(double w, double h, size_t id)
{
	return new (detail::get_node_pool()->alloc()) node_t(w, h, id);
}

void acg_dev_t::free_node(node_t *node)
{
	node->~node_t();

	detail::get_node_pool()->free(node);
}

void acg_dev_t::insert(size_t where, node_t *node)
{
	assert(where <= nodes_.size());

	nodes_.insert(nodes_.begin()+where, node);
}

void acg_dev_t::erase(size_t where)
{
	assert(where < nodes_.size());

	assert(nodes_[where]->edges(COLOR_RED)->empty(POS_IN));
	assert(nodes_[where]->edges(COLOR_RED)->empty(POS_OUT));
	assert(nodes_[where]->edges(COLOR_BLACK)->empty(POS_IN));
	assert(nodes_[where]->edges(COLOR_BLACK)->empty(POS_OUT));

	nodes_.erase(nodes_.begin()+where);
}

void acg_dev_t::push(node_t *node, pos_t pos)
{
	if (pos == POS_IN)
		insert(0, node);
	else
		nodes_.push_back(node);
}

void acg_dev_t::pop(pos_t pos)
{
	if (pos == POS_IN)
		erase(0);
	else
	{
		assert(!nodes_.empty());

		assert(nodes_.back()->edges(COLOR_RED)->empty(POS_IN));
		assert(nodes_.back()->edges(COLOR_RED)->empty(POS_OUT));
		assert(nodes_.back()->edges(COLOR_BLACK)->empty(POS_IN));
		assert(nodes_.back()->edges(COLOR_BLACK)->empty(POS_OUT));
		
		nodes_.pop_back();
	}
}

node_t *acg_dev_t::front(pos_t pos)
{
	assert(num_nodes() > 0);

	if (pos == POS_IN)
		return nodes_.front();
	else
		return nodes_.back();
}

void acg_dev_t::free_edges(edge_t *first, edge_t *last, pos_t pos)
{
	assert(first->valid());
	assert(last->valid());

	assert(first->node(pos) == last->node(pos));

	for (; first != last;)
	{
		edge_t *tmp = first; first = first->next(pos);

		edge_list_t::erase(tmp, POS_IN);
		edge_list_t::erase(tmp, POS_OUT);

		free_edge(tmp);
	}
}

void acg_dev_t::free_edges(node_t *node, pos_t pos)
{
	free_edges(node, COLOR_RED, pos);
	free_edges(node, COLOR_BLACK, pos);
}

void acg_dev_t::free_edges(node_t *node, color_t c)
{
	free_edges(node, c, POS_IN);
	free_edges(node, c, POS_OUT);
}

void acg_dev_t::free_edges(node_t *node, color_t c, pos_t pos)
{
	edge_list_t *el = node->edges(c);

	free_edges(el->begin(pos), el->end(pos), pos);
}

void acg_dev_t::free_all_edges()
{
	if (n_edges_ != 0)
		for (size_t i = 0; i < num_nodes(); ++i)
			free_edges(nodes_[i], POS_IN);
	assert(n_edges_ == 0);
}

void acg_dev_t::duplicate_edges(const acg_base_t &src)
{
	assert(num_nodes() == src.num_nodes());

	const size_t n = num_nodes();

	// clear all the dest nodes first
	for (size_t i = 0; i < n; ++i)
	{
		assert(at(i)->valid());
		assert(src[i]->valid());

		assert(at(i)->edges(COLOR_RED)->empty(POS_IN));
		assert(at(i)->edges(COLOR_RED)->empty(POS_OUT));
		assert(at(i)->edges(COLOR_BLACK)->empty(POS_IN));
		assert(at(i)->edges(COLOR_BLACK)->empty(POS_OUT));

		src[i]->m_data = at(i);
	}

	// create edges and link them at POS_IN side
	for (size_t from = 0; from < n; ++from)
	{
		dup_edges_POS_IN(COLOR_RED, src[from]);
		dup_edges_POS_IN(COLOR_BLACK, src[from]);
	}

	// link edges at POS_OUT side
	for (size_t to = 0; to < n; ++to)
	{
		dup_edges_POS_OUT(COLOR_RED, src[to]);
		dup_edges_POS_OUT(COLOR_BLACK, src[to]);
	}

	assert(num_edges() == src.num_edges());
}

void acg_dev_t::dump_cg_edges(FILE *fp) const
{
	const size_t n = num_nodes();

	for (size_t i = 0; i < n; ++i)
		at(i)->m_data = reinterpret_cast<void *>(i);

	fprintf(fp, "# dumping cg...............................................\n");
	fprintf(fp, "nodes %u, edges %u\n", num_nodes(), num_edges());

	for (size_t j = 0; j < n; ++j)
		fprintf(fp,
			"node %p, id %u, order %u, width %.0f, height %.0f\n",
			at(j), at(j)->m_id,
			reinterpret_cast<size_t>(at(j)->m_data),
			at(j)->width(COLOR_RED), at(j)->width(COLOR_BLACK));

	for (size_t k = 0; k < n; ++k)
	{
		dump_node_edges(at(k), COLOR_RED, POS_IN, fp);
		dump_node_edges(at(k), COLOR_RED, POS_OUT, fp);
		dump_node_edges(at(k), COLOR_BLACK, POS_IN, fp);
		dump_node_edges(at(k), COLOR_BLACK, POS_OUT, fp);
	}

	fprintf(fp, "\n");
}

void acg_dev_t::update_order(const size_t order[])
{
	const size_t n = num_nodes();
	std::vector<node_t *> nodes(n);

	for (size_t i = 0; i < n; ++i)
		nodes[i] = at(order[i]);

	nodes_.swap(nodes);
}

void acg_dev_t::dup_edges_POS_IN(color_t c, const node_t *src)
{
	for (edge_list_t::const_range_t r = src->edges(c)->all(POS_IN);
		r.first != r.second; r.first = r.first->next(POS_IN))
	{
		edge_t *new_edge = alloc_edge();

		r.first->m_data = new_edge;

		reinterpret_cast<node_t *>(src->m_data)
			->edges(c)->push_back(new_edge, POS_IN);
	}
}

void acg_dev_t::dup_edges_POS_OUT(color_t c, const node_t *src)
{
	for (edge_list_t::const_range_t r = src->edges(c)->all(POS_OUT);
		r.first != r.second; r.first = r.first->next(POS_OUT))
	{
		edge_t *new_edge = reinterpret_cast<edge_t *>(r.first->m_data);

		reinterpret_cast<node_t *>(src->m_data)
			->edges(c)->push_back(new_edge, POS_OUT);
	}
}

void acg_dev_t::dump_node_edges(const node_t *node,
	color_t c, pos_t pos, FILE *fp) const
{
	const char *color = (c == COLOR_RED)? "red": "black";

	size_t node_id = reinterpret_cast<size_t>(node->m_data);
	
	size_t last_id = node_id;
	
	for (edge_list_t::const_range_t r = node->edges(c)->all(pos);
		r.first != r.second; r.first = r.first->next(pos))
	{
		const node_t *next = r.first->node(other(pos));

		size_t next_id = reinterpret_cast<size_t>(next->m_data);

		assert(node == r.first->node(pos));

		assert((pos == POS_IN)? (last_id < next_id): (last_id > next_id));
		
		last_id = next_id;

		fprintf(fp, "edge %u->%u, color %s\n", node_id, next_id, color);
	}
}

} // namespace acg
