/*
 * Copyright (c) 2004-2005 Northwestern University.
 *                         NuCAD Group.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * acg_dev.h - Internal operations on acg_base_t.
 *
 * Authors: Hai Zhou, haizhou@ece.northwestern.edu
 *          Jia Wang, jwa112@ece.northwestern.edu
 */

#ifndef ACG_DEV_H
#define ACG_DEV_H

#include "acg_base.h"

namespace acg
{

class acg_dev_t
{
	size_t &n_edges_;
	std::vector<node_t *> &nodes_;

	// helper functions

	void dup_edges_POS_IN(color_t c, const node_t *src);
	void dup_edges_POS_OUT(color_t c, const node_t *src);

	void dump_node_edges(const node_t *node, color_t c, pos_t pos, FILE *fp) const;

public:
	explicit acg_dev_t(acg_base_t &base)
		: n_edges_(base.n_edges_), nodes_(base.nodes_)
		{}

	// statistics
	size_t num_nodes() const {return nodes_.size();}
	size_t num_edges() const {return n_edges_;}
	bool empty() const {return nodes_.empty();}

	// individual node visiting
	// be aware you can even modify them
	node_t *&at(size_t i) {assert(i < num_nodes()); return nodes_[i];}
	const node_t *at(size_t i) const {assert(i < num_nodes()); return nodes_[i];}
	node_t *&operator[](size_t i) {return at(i);}
	const node_t *operator[](size_t i) const {return at(i);}

	// we count # of edges here
	edge_t *alloc_edge();
	void free_edge(edge_t *edge);

	// free a group of edges

	// free [first, last) at pos side
	void free_edges(edge_t *first, edge_t *last, pos_t pos);

	// free edges associated with a node
	void free_edges(node_t *node, pos_t pos);
	void free_edges(node_t *node, color_t c);
	void free_edges(node_t *node, color_t c, pos_t pos);

	void free_all_edges();

	// allocator for nodes
	node_t *alloc_node();
	node_t *alloc_node(double w, double h, size_t id);
	void free_node(node_t *node);

	// insert/erase nodes

	// POS_IN: front, POS_OUT: back
	node_t *front(pos_t pos);
	void push(node_t *node, pos_t pos);
	void pop(pos_t pos);

	void insert(size_t where, node_t *node);
	void erase(size_t where);

	// update nodes order
	// order[0] is the old pos of the new 0
	void update_order(const size_t order[]);

	// dumping for cg (acg/racg/rcg) debugging use
	// node_t::m_data will be overwrited
	// see acg::xml for better dumping
	void dump_cg_edges(FILE *fp) const;

	// duplicate edges
	// node_t::m_data and edge_t::m_data will be overwrited
	// assume dest has the same number of nodes as src and 0 edges
	void duplicate_edges(const acg_base_t &src);

}; // class acg_dev_t

} // namespace acg

#endif // ACG_DEV_H
