/*
 * Copyright (c) 2004-2005 Northwestern University.
 *                         NuCAD Group.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * acg_racg.h - Implement algorithms for ACG and Reduced ACG.
 *
 * Authors: Hai Zhou, haizhou@ece.northwestern.edu
 *          Jia Wang, jwa112@ece.northwestern.edu
 */

#include "../misc/common.h"

#include "acg_racg.h"
#include "acg_dev.h"

namespace acg
{

namespace detail
{

void racg_create_edges(acg_dev_t &base, node_t *node, node_t *next, color_t c, pos_t pos);

edge_t *racg_relink_remove(acg_dev_t &base, edge_list_t *edges, edge_t *AB, pos_t pos);

color_t node_color(const node_t *node, pos_t pos);

void racg_erase_case1(acg_dev_t &base, node_t *node, color_t c);
void racg_erase_case2(acg_dev_t &base, node_t *node, color_t out);

} // namespace detail

void racg_build(acg_base_t &b)
{
	using detail::racg_create_edges;

	acg_dev_t base(b);

	for (size_t i = 1; i < base.num_nodes(); ++i)
		racg_create_edges(base, base[i], base[i-1], color_t(rand()%2), POS_OUT);
}

void racg_push(acg_base_t &b,
	double w, double h, size_t id,
	color_t c, pos_t pos)
{
	using detail::racg_create_edges;

	acg_dev_t base(b);

	node_t *node = base.alloc_node(w, h, id);

	if (!base.empty())
		racg_create_edges(base, node, base.front(pos), c, pos);

	base.push(node, pos);
}

void racg_pop(acg_base_t &b, pos_t pos)
{
	acg_dev_t base(b);

	node_t *tmp = base.front(pos);

	base.free_edges(tmp, pos);

	base.pop(pos);

	base.free_node(tmp);
}

/*
 * refer to erase case 1
 *
 * X--A--B--C
 * X  :  :  :
 * X--DDDDDDD--Y <= X--A--B--CCCCCCC
 *    :  :  :  Y       :  :  :  :  :
 *    E--F--G--Y       EEEEEEE--F--G--Y
 */
void racg_insert(acg_base_t &b, size_t where,
	double w, double h, size_t id)
{
	using detail::node_color;

	if (where == 0)
	{
		racg_push(b, w, h, id, color_t(rand()%2), POS_IN);
		return;
	}
	if (where == b.num_nodes())
	{
		racg_push(b, w, h, id, color_t(rand()%2), POS_OUT);
		return;
	}

	acg_dev_t base(b);

	node_t *node = base.alloc_node(w, h, id);

	node_t *nD = node;
	node_t *nE = base[where];

	color_t cCE = node_color(nE, POS_OUT);
	edge_t *eCE = nE->edges(cCE)->front(POS_OUT);
	
	for (int i = 0; i < 2; ++i)
	{
		pos_t pos = pos_t(i);

		// following variables named according to pos == POS_IN

		node_t *nE = eCE->node(other(pos));

		// create D->E
		edge_t *eDE = base.alloc_edge();
		nD->edges(cCE)->push_front(eDE, pos);
		nE->edges(cCE)->push_front(eDE, other(pos));

		for (edge_t *eCF = eCE->next(pos);
			eCF != eCE->node(pos)->edges(cCE)->end(pos);
			eCF = eCE->next(pos)) // C->F, C->G, etc
		{
			if (rand()%2 == 0)
				break;

			// C->F becomes D->F
			edge_list_t::erase(eCF, pos);
			nD->edges(cCE)->push_back(eCF, pos);
		}

		// create D->Y if needed
		node_t *nG = nD->farest_node(cCE, pos);
		assert(nG != nD);
		if (!nG->edges(other(cCE))->empty(pos))
		{
			edge_t *eGY = nG->edges(other(cCE))->front(pos);

			edge_t *eDY = base.alloc_edge();
			nD->edges(other(cCE))->push_front(eDY, pos);
			edge_list_t::insert(eGY->next(other(pos)), eDY, other(pos));
		}
	}
	
	// remove C->E
	edge_list_t::erase(eCE, POS_IN);
	edge_list_t::erase(eCE, POS_OUT);
	base.free_edge(eCE);
	
	// insert to base
	base.insert(where, node);
}

void racg_erase(acg_base_t &b, size_t where)
{
	using detail::node_color;
	using detail::racg_erase_case1;
	using detail::racg_erase_case2;

	if (where == 0)
	{
		racg_pop(b, POS_IN);
		return;
	}
	if (where == b.num_nodes()-1)
	{
		racg_pop(b, POS_OUT);
		return;
	}

	acg_dev_t base(b);

	node_t *node = base[where];

	color_t in = node_color(node, POS_IN);
	color_t out = node_color(node, POS_OUT);

	if (in == out)
		racg_erase_case1(base, node, in);
	else
		racg_erase_case2(base, node, out);

	base.erase(where);

	base.free_node(node);
}

/*
 *          S
 *          :                      SSSSSSSSSS
 * X--A--B--CCCCCCC--Y             :  :  :  :
 * X  :  :  :  :  :  Y => X--A--B--C--D--E--F--Y
 * X--DDDDDDD--E--F--Y       :  :  :  :
 *          :                TTTTTTTTTT
 *          T
 */
void racg_switch_adjacent(acg_base_t &b, size_t where, pos_t pos)
{
	using detail::node_color;
	using detail::racg_relink_remove;

	acg_dev_t base(b);

	node_t *node = base[where];
	color_t cCD = node_color(node, pos);

	if (node->edges(cCD)->empty(pos))
		return;

	edge_t *eCD = node->edges(cCD)->front(pos);
	node_t *nC = eCD->node(POS_IN);
	node_t *nD = eCD->node(POS_OUT);

	edge_t *eSE = racg_relink_remove(base, nC->edges(cCD), eCD, POS_IN);
	edge_t *eBT = racg_relink_remove(base, nD->edges(cCD), eCD, POS_OUT);

/*
 *          SSSSSSS
 *          :  :  :
 * X--A--B--C--------Y
 * X  :  :  :  :  :  Y
 * X--------D--E--F--Y
 *    :  :  :
 *    TTTTTTT
 */

	edge_list_t *otherC = nC->edges(other(cCD));
	edge_list_t *otherD = nD->edges(other(cCD));

	// remove alter edges
	if (!otherC->empty(POS_IN))
	{
		edge_t *eCY = otherC->pop_front(POS_IN);
		edge_list_t::erase(eCY, POS_OUT);
		base.free_edge(eCY);
	}
	if (!otherD->empty(POS_OUT))
	{
		edge_t *eXD = otherD->pop_front(POS_OUT);
		edge_list_t::erase(eXD, POS_IN);
		base.free_edge(eXD);
	}

	// remove adjacent edge
	edge_list_t::erase(eCD, POS_IN);
	edge_list_t::erase(eCD, POS_OUT);

/*
 *          SSSSSSSSSS
 *          :     :  :
 * X--A--B--C  D--E--F--Y
 *    :  :     :
 *    TTTTTTTTTT
 */

	assert(nC->edges(cCD)->empty(POS_IN));
	assert(otherC->empty(POS_IN));
	assert(nD->edges(cCD)->empty(POS_OUT));
	assert(otherD->empty(POS_OUT));

	// new adjacent edge, using the other color
	otherC->push_back(eCD, POS_IN); // to an empty list
	otherD->push_back(eCD, POS_OUT); // to an empty list

	// add two more edges
	if (eSE != 0)
	{
		edge_t *eSD = base.alloc_edge();
		edge_list_t::insert(eSE, eSD, POS_IN);
		nD->edges(cCD)->push_back(eSD, POS_OUT); // to an empty list
	}
	if (eBT != 0)
	{
		edge_t *eCT = base.alloc_edge();
		edge_list_t::insert(eBT, eCT, POS_OUT);
		nC->edges(cCD)->push_back(eCT, POS_IN); // to an empty list
	}
}

/*
 * S             SSSSSSS
 * :             :     :
 * AAAAAAA--E => AAAA--D
 * :  :  :  E    :  :  D
 * B--C--D--E    B--C--D--E
 */
void racg_switch_last(acg_base_t &b, size_t where, pos_t pos)
{
	using detail::node_color;

	acg_dev_t base(b);

	node_t *nA = base[where];
	color_t cAB = node_color(nA, pos);
	edge_list_t *edgeA = nA->edges(cAB);

	if (edgeA->empty(pos)) // no such edge, it is the last one
		return;
	
	edge_t *eAD = edgeA->back(pos);
	if (edgeA->front(pos) == eAD) // should call swap_adjacent
		return;

	node_t *nD = eAD->node(other(pos));

	edge_list_t::erase(eAD, POS_IN);
	edge_list_t::erase(eAD, POS_OUT);

	assert(nD->edges(cAB)->empty(other(pos)));

	edge_list_t *otherA = nA->edges(other(cAB));
	if (!otherA->empty(pos))
	{
		edge_t *eAE = otherA->pop_front(pos);
		edge_list_t::erase(eAE, other(pos));
		base.free_edge(eAE);
	}
	assert(otherA->empty(pos));

	otherA->push_back(eAD, pos);
	nD->edges(other(cAB))->push_back(eAD, other(pos));

/*
 * S
 * :
 * AAAA--D
 * :  :  D
 * B--C--D--E
 */

	if (!edgeA->empty(other(pos)))
	{
		edge_t *eSA = edgeA->front(other(pos));

		edge_t *eSD = base.alloc_edge();
		edge_list_t::insert(eSA->next(pos), eSD, pos);
		nD->edges(cAB)->push_back(eSD, other(pos)); // to an empty list
	}
}

/*
 * C--D--E--F--T => C--D--E--F--T
 * :  :  :  F       :  :  :  :  T
 * BBBBBBB--F       BBBBBBBBBB--T
 * :        F       :           T
 * A--------F       A-----------T
 * :        :       :
 * XXXXXXXXXX       X
 */
void racg_switch_alter(acg_base_t &b, size_t where, pos_t pos)
{
	using detail::node_color;
	using detail::racg_relink_remove;

	acg_dev_t base(b);

	node_t *nB = base[where];
	color_t cBC = node_color(nB, pos);
	edge_list_t *otherB = nB->edges(other(cBC));

	if (otherB->empty(pos)) // no such edge
		return;

	edge_t *eBF = otherB->front(pos);
	node_t *nF = eBF->node(other(pos));
	
	edge_t *eAT = racg_relink_remove(base, nF->edges(other(cBC)), eBF, other(pos));

/*
 * C--D--E--F--T
 * :  :  :  F  T
 * BBBBBBB--F  T
 * :        :  T
 * A-----------T
 * :        :
 * XXXXXXXXXX
 */

	edge_list_t *edgeF = nF->edges(cBC);

	// remove alter edges
	if (!edgeF->empty(other(pos)))
	{
		edge_t *eXF = edgeF->pop_front(other(pos));
		edge_list_t::erase(eXF, pos);
		base.free_edge(eXF);
	}

	edge_list_t::erase(eBF, POS_IN);
	edge_list_t::erase(eBF, POS_OUT);

/*
 * C--D--E--F--T
 * :  :  :     T
 * BBBBBBB     T
 * :           T
 * A-----------T
 * :
 * X
 */

	assert(edgeF->empty(other(pos)));
	assert(otherB->empty(pos));

	// relink eBF using the other color
	nB->edges(cBC)->push_back(eBF, pos); // must be the last one
	edgeF->push_back(eBF, other(pos)); // to an empty list

	// add one more edge
	if (eAT != 0)
	{
		edge_t *eBT = base.alloc_edge();
		otherB->push_back(eBT, pos); // to an empty list
		edge_list_t::insert(eAT, eBT, other(pos));
	}
}

void racg_to_acg(acg_base_t &b, racg_acg_context_t &ctx)
{
	using detail::node_color;

	assert(ctx.empty());

	acg_dev_t base(b);

	if (base.num_nodes() < 2)
		return;

	ctx.reserve(2*base.num_nodes());

	for (size_t i = 1; i < base.num_nodes(); ++i)
	{
		node_t *node = base[i];
		color_t c = node_color(node, POS_OUT);
		edge_list_t *edges = node->edges(c);
		edge_list_t *others = node->edges(other(c));

		if (others->empty(POS_OUT))
			continue;

		assert(others->front(POS_OUT) == others->back(POS_OUT)); // iacg

		edge_t *bridge = others->front(POS_OUT)->prev(POS_IN);

		for (;;)
		{
			assert(bridge->node(POS_OUT) == edges->back(POS_OUT)->node(POS_IN));

			bridge = bridge->next(POS_OUT);
			if (bridge == bridge->node(POS_OUT)->edges(other(c))->end(POS_OUT))
				break;

			// add acg edge
			edge_t *acg_edge = base.alloc_edge();
			bridge->node(POS_IN)->edges(c)->push_back(acg_edge, POS_IN);
			edges->push_back(acg_edge, POS_OUT);

			ctx.push_back(acg_edge);

			// go next
			c = other(c);
			std::swap(edges, others);
			bridge = acg_edge->prev(POS_IN);
		}
	}
}

void acg_to_racg(acg_base_t &b, racg_acg_context_t &ctx)
{
	acg_dev_t base(b);

	for (size_t i = 0; i < ctx.size(); ++i)
	{
		edge_list_t::erase(ctx[i], POS_IN);
		edge_list_t::erase(ctx[i], POS_OUT);
		base.free_edge(ctx[i]);
	}
	ctx.clear();
}

namespace detail
{

void racg_create_edges(acg_dev_t &base, node_t *node, node_t *next, color_t c, pos_t pos)
{
	node_t *prev = 0;
	for (color_t cur = c; prev != next;
		prev = next, next = next->nearest_node(other(c), pos))
	{
		assert(next->edges(c)->empty(other(pos))); // Reduced ACG

		edge_t *edge = base.alloc_edge();
		node->edges(cur)->push_back(edge, pos);
		next->edges(cur)->push_back(edge, other(pos));
		
		if (cur != c)
			return; // alter edge
		else
			cur = color_t(rand()%2);
	}
}

/*
 * S                SSSSSSSSSSSSS
 * :                :  :  :  :  :
 * AAAAAAAAAAAAA => A  :  :  :  :
 * :  :  :  :  :    :  :  :  :  :
 * B--C--D--E--F    B--C--D--E--F
 */
edge_t *racg_relink_remove(acg_dev_t &base, edge_list_t *edges, edge_t *AB, pos_t pos)
{
	edge_t *AC = AB->next(pos);

	if (edges->empty(other(pos)))
		// remove
	{
		base.free_edges(AC, edges->end(pos), pos);

		return 0;
	}
	else
		// relink
	{
		edge_t *SA = edges->front(other(pos));
		edge_list_t::splice(SA->next(pos), AC, edges->end(pos), pos);

		return SA->next(pos);
	}
}

color_t node_color(const node_t *node, pos_t pos)
{
	const node_t *red = node->nearest_node(COLOR_RED, pos);

	if (red == node)
		return COLOR_BLACK;

	if (red->nearest_node(COLOR_RED, other(pos)) == node)
		return COLOR_RED;
	else
		return COLOR_BLACK;
}

/*
 * case 1
 *
 * X--A--B--C
 * X  :  :  :
 * X--DDDDDDD--Y => X--A--B--CCCCCCC
 *    :  :  :  Y       :  :  :  :  :
 *    E--F--G--Y       EEEEEEE--F--G--Y
 */
void racg_erase_case1(acg_dev_t &base, node_t *node, color_t c)
{
	node_t *nD = node;

	edge_list_t *edgeD = nD->edges(c);
	edge_list_t *otherD = nD->edges(other(c));

	// remove X->D and D->Y
	if (!otherD->empty(POS_IN))
	{
		edge_t *eDY = otherD->pop_front(POS_IN);
		edge_list_t::erase(eDY, POS_OUT);
		base.free_edge(eDY);
	}
	if (!otherD->empty(POS_OUT))
	{
		edge_t *eXD = otherD->pop_front(POS_OUT);
		edge_list_t::erase(eXD, POS_IN);
		base.free_edge(eXD);
	}

/*
 * X--A--B--C
 *    :  :  :
 *    DDDDDDD
 *    :  :  :
 *    E--F--G--Y
 */

	assert(otherD->empty(POS_IN) && otherD->empty(POS_OUT));

	edge_t *eCD = edgeD->front(POS_OUT);
	edge_t *eDE = edgeD->front(POS_IN);

	// D->F, D->G, etc become C->F, C->G, etc
	edge_list_t::splice(eCD->next(POS_IN),
		eDE->next(POS_IN), edgeD->end(POS_IN), POS_IN);
	// B->D, A->D, etc become B->E, A->E, etc
	edge_list_t::splice(eDE->next(POS_OUT),
		eCD->next(POS_OUT), edgeD->end(POS_OUT), POS_OUT);

/*
 * X--A--B--CCCCCCC
 *    :  :  :  :  :
 *    :  :  D  :  :
 *    :  :  :  :  :
 *    EEEEEEE--F--G--Y
 */

	// reuse C->D to be C->E
	edge_list_t::erase(eCD, POS_OUT);
	edge_list_t::insert(eDE, eCD, POS_OUT);

	// free D->E
	edge_list_t::erase(eDE, POS_IN);
	edge_list_t::erase(eDE, POS_OUT);
	base.free_edge(eDE);

	assert(edgeD->empty(POS_IN));
	assert(edgeD->empty(POS_OUT));
}

/*
 * case 2
 *
 * X--A--B--CCCC    X--A--B--C
 * X  :  :  :  :    X  :  :  :
 * X--DDDDDDD--E => X--EEEEEEE
 *    DDDDDDD  :    X        :
 *    DDDDDDD--F    X--------F
 *    DDDDDDD  :    X        :
 *    DDDDDDD--G    X--------G
 *    :        :             :
 *    YYYYYYYYYY             Y
 */
void racg_erase_case2(acg_dev_t &base, node_t *node, color_t out)
{
	color_t cCD = out;
	pos_t pos = POS_OUT;
	node_t *nD = node;
	node_t *nC = nD->nearest_node(cCD, pos);

	if (nC->nearest_node(cCD, other(pos)) == nC->farest_node(cCD, other(pos)))
	{
		cCD = other(cCD);
		pos = POS_IN;
	}

	edge_list_t *edges = nD->edges(cCD);
	edge_list_t *other_edges = nD->edges(other(cCD));

	// free D->Y
	if (!edges->empty(other(pos)))
	{
		edge_t *eDY = edges->pop_front(other(pos));
		edge_list_t::erase(eDY, pos);
		base.free_edge(eDY);
	}

	edge_t *eCD = edges->front(pos);
	edge_t *eCE = eCD->next(other(pos));

	// B->D, A->D, etc become B->E, A->E, etc
	edge_list_t::splice(eCE->next(pos), eCD->next(pos), edges->end(pos), pos);

	// free C->D
	edge_list_t::erase(eCD, POS_IN);
	edge_list_t::erase(eCD, POS_OUT);
	base.free_edge(eCD);

	assert(edges->empty(POS_IN));
	assert(edges->empty(POS_OUT));

	edge_list_t::range_t r = other_edges->all(other(pos));
	if (other_edges->empty(pos))
		// no X, free D->E, D->F, D->G, etc
		base.free_edges(r.first, r.second, other(pos));
	else
	{
		edge_t *eXD = other_edges->front(pos);

		// has X, D->E, D->F, D->G, etc become X->E, X->F, X->G, etc
		edge_list_t::splice(eXD->next(other(pos)),
			r.first, r.second, other(pos));

		// free X->D
		edge_list_t::erase(eXD, POS_IN);
		edge_list_t::erase(eXD, POS_OUT);
		base.free_edge(eXD);
	}
	assert(other_edges->empty(POS_IN));
	assert(other_edges->empty(POS_OUT));
}

} // namespace detail

} // namespace acg
