/*
 * Copyright (c) 2004-2005 Northwestern University.
 *                         NuCAD Group.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * cg_pack.cpp - Longest path based packing.
 *
 * Authors: Hai Zhou, haizhou@ece.northwestern.edu
 *          Jia Wang, jwa112@ece.northwestern.edu
 */

#include "../misc/common.h"

#include "cg_pack.h"

namespace acg
{

namespace detail
{

void nodes_bb_reset(acg_base_t &base, pos_t pos);

misc::dbl_pair packing_longest_path_impl(acg_base_t &base, pos_t pos, size_t &count);

size_t longest_path_propagate(node_t *node, pos_t pos, color_t c, double &bound);

void nodes_bb_reset_unsort(acg_base_t &base, color_t c);

double calc_boundary_unsort(node_t *node, color_t c);

} // namespace detail

misc::dbl_pair packing_longest_path(acg_base_t &base, size_t &count)
{
	return detail::packing_longest_path_impl(base, POS_IN, count);
}

misc::dbl_pair packing_longest_path(acg_base_t &base)
{
	size_t count = 0;
	return detail::packing_longest_path_impl(base, POS_IN, count);
}

void longest_path_bb(acg_base_t &base, misc::dbl_pair shape)
{
	size_t count = 0;

	detail::packing_longest_path_impl(base, POS_OUT, count);

	const size_t n = base.num_nodes();
	for (size_t i = 0; i < n; ++i)
	{
		base[i]->bb(COLOR_RED, POS_OUT)
			= shape.first-base[i]->bb(COLOR_RED, POS_OUT);
		base[i]->bb(COLOR_BLACK, POS_OUT)
			= shape.second-base[i]->bb(COLOR_BLACK, POS_OUT);
	}
}

double packing_lp_unsort(acg_base_t &base, color_t c)
{
	detail::nodes_bb_reset_unsort(base, c);

	const size_t n = base.num_nodes();
	double width = 0;

	for (size_t i = 0; i < n; ++i)
	{
		double t = detail::calc_boundary_unsort(base[i], c);
		if (width < t)
			width = t;
	}
	return width;
}

namespace detail
{

void nodes_bb_reset(acg_base_t &base, pos_t pos)
{
	const size_t n = base.num_nodes();

	for (size_t i = 0; i < n; ++i)
		base[i]->bb(COLOR_RED, pos) = base[i]->bb(COLOR_BLACK, pos) = 0;
}

size_t longest_path_propagate(node_t *node, pos_t pos, color_t c, double &bound)
{
	size_t count = 0;

	double t = node->bb(c, pos)+node->width(c);

	if (node->edges(c)->empty(pos))
	{
		if (bound < t)
			bound = t;
	}
	else
		for (edge_list_t::range_t r = node->edges(c)->all(pos);
			r.first != r.second; r.first = r.first->next(pos))
		{
			node_t *to = r.first->node(other(pos));
            
			if (to->bb(c, pos) < t)
				to->bb(c, pos) = t;

			++count;
		}

	return count;
}

misc::dbl_pair packing_longest_path_impl(acg_base_t &base, pos_t pos, size_t &count)
{
	const size_t n = base.num_nodes();

	nodes_bb_reset(base, pos);

	double width = 0, height = 0;

	if (pos == POS_IN)
		for (size_t i = 0; i < n; ++i)
		{
			count += longest_path_propagate(base[i], pos, COLOR_RED, width);
			count += longest_path_propagate(base[i], pos, COLOR_BLACK, height);
		}
	else
		for (size_t i = n; i > 0; --i)
		{
			count += longest_path_propagate(base[i-1], pos, COLOR_RED, width);
			count += longest_path_propagate(base[i-1], pos, COLOR_BLACK, height);
		}

	return std::make_pair(width, height);
}

void nodes_bb_reset_unsort(acg_base_t &base, color_t c)
{
	const size_t n = base.num_nodes();

	for (size_t i = 0; i < n; ++i)
		base[i]->bb(c, POS_IN) = base[i]->bb(c, POS_OUT) = -1;
}

double calc_boundary_unsort(node_t *node, color_t c)
{
	if (node->bb(c, POS_IN) < -0.5)
		// -1 is always less than and 0 is always greater than -0.5
	{
		node->bb(c, POS_IN) = 0;

		for (edge_list_t::range_t r = node->edges(c)->all(POS_OUT);
			r.first != r.second; r.first = r.first->next(POS_OUT))
		{
			double t = calc_boundary_unsort(r.first->node(POS_IN), c);
			
			if (node->bb(c, POS_IN) < t)
				node->bb(c, POS_IN) = t;
		}
	}
	return node->boundary(c, POS_IN);
}

} // namespace detail

} // namespace acg
