/*
 * Copyright (c) 2004-2005 Northwestern University.
 *                         NuCAD Group.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * fplan.cpp - Implement interface for floorplanning.
 *
 * Authors: Hai Zhou, haizhou@ece.northwestern.edu
 *          Jia Wang, jwa112@ece.northwestern.edu
 *
 */

#include "../misc/common.h"

#include "fplan.h"

namespace fplan
{

namespace detail
{

class area_cost_t : public cost_i
{
	fplan_i *fp_;

	virtual double cost()
	{
		misc::dbl_pair shape = fp_->get_shape(); // eval
		return shape.first*shape.second;
	}

	virtual void release() {delete this;}

public:
	area_cost_t(fplan_i *fp) : fp_(fp) {}
}; // class area_cost_t

class fixed_cost_t : public cost_i
{
	fplan_i *fp_;
	double w_, h_;
	bool rotate_;

	virtual double cost()
	{
		misc::dbl_pair shape = fp_->get_shape(); // eval
			
		double w = shape.first, h = shape.second;

		if (rotate_ && (w < h))
			std::swap(w, h);

		w = (w > w_)? w-w_: 0;
		h = (h > h_)? h-h_: 0;

		return exp(h/h_+w/w_);
	}

	virtual void release() {delete this;}

public:
	fixed_cost_t(fplan_i *fp, double w, double h, bool rotate)
		: fp_(fp), w_(w), h_(h), rotate_(rotate)
	{
		if (rotate_ && (w_ < h_))
			std::swap(w_, h_);
	}
}; // class fixed_cost_t

class square_cost_t : public cost_i
{
	fplan_i *fp_;

	virtual double cost()
	{
		misc::dbl_pair shape = fp_->get_shape(); // eval
		
		double ar = (shape.first < shape.second)?
			shape.second/shape.first-1:
			shape.first/shape.second-1;

		return exp(ar);
	}

	virtual void release() {delete this;}

public:
	square_cost_t(fplan_i *fp) : fp_(fp) {}
}; // class square_cost_t

class mnets_cost_t : public cost_i
{
	fplan_i *fp_;

	const module_nets_t &mnets_;

	typedef double (module_nets_t::*WL_func_t)(const misc::dbl_pair centers[]) const;
	WL_func_t wl_func_;

	const size_t nm_, np_;
	const misc::dbl_pair *pins_;

	std::vector<misc::dbl_pair> centers_;

	virtual double cost()
	{
		// you should create a new cost_ptr if the floorplan changed
		assert(nm_ == fp_->get_num_modules());

		// get shape and centers
		misc::dbl_pair shape = fp_->get_centers(0, &centers_[0]);

		// compute pin positions
		for (size_t i = 0; i < np_; ++i)
			centers_[i+nm_] = pins_[i]*shape;

//		for (size_t j = 0; j < centers_.size(); ++j)
//			printf("%u, %.0f, %.0f\n", j, centers_[j].first, centers_[j].second);

		return (mnets_.*wl_func_)(&centers_[0]);
	}

	virtual void release() {delete this;}

public:
	mnets_cost_t(fplan_i *fp, const module_nets_t &mnets,
		WL_func_t wl_func, size_t np, const misc::dbl_pair pins[])
		: fp_(fp), mnets_(mnets), wl_func_(wl_func),
		nm_(fp->get_num_modules()), np_(np), pins_(pins),
		centers_(fp->get_num_modules()+np)
		{}
}; // class mnets_cost_t

} // namespace detail

cost_ptr fplan_i::area_cost()
{
	return cost_ptr(new detail::area_cost_t(this));
}

cost_ptr fplan_i::square_cost()
{
	return cost_ptr(new detail::square_cost_t(this));
}

cost_ptr fplan_i::fixed_cost(double w, double h, bool rotate)
{
	return cost_ptr(new detail::fixed_cost_t(this, w, h, rotate));
}

cost_ptr fplan_i::HPWL_cost(const module_nets_t &mnets,
	size_t np, const misc::dbl_pair pins[])
{
	return cost_ptr(new detail::mnets_cost_t(this, mnets, &module_nets_t::HPWL, np, pins));
}

cost_ptr fplan_i::QPWL_cost(const module_nets_t &mnets,
	size_t np, const misc::dbl_pair pins[])
{
	return cost_ptr(new detail::mnets_cost_t(this, mnets, &module_nets_t::QPWL, np, pins));
}

} // namespace fplan
