/*
 * Copyright (c) 2004-2005 Northwestern University.
 *                         NuCAD Group.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * fplan_rcg.cpp - Implement RCG fploorplanning.
 *
 * Authors: Hai Zhou, haizhou@ece.northwestern.edu
 *          Jia Wang, jwa112@ece.northwestern.edu
 *
 */

#include "../misc/common.h"

#include "fplan_rcg.h"

#include "../acg/cg_pack.h"
#include "../acg/rcg.h"

namespace fplan
{

class rcg_perturb_t : public detail::perturb_i
{
protected:
	rcg_fp_t *fp_;
	double rotateP_, exchangeP_;

	acg::acg_base_t best_;
	std::vector<size_t> best_rotated_;

	int choice_;
	size_t n1_, n2_;

	virtual void perturbation(double T);
	virtual void reject();
	virtual void record_best();
	virtual void swap_best();

	// you should use the shared_ptr interface
	virtual void release() {delete this;}
	~rcg_perturb_t() {}

public:
	rcg_perturb_t(rcg_fp_t *fp, double rotateP, double exchangeP);

}; // class rcg_perturb_t


rcg_fp_t::rcg_fp_t()
	: mapped_(false)
{
}

rcg_fp_t::rcg_fp_t(size_t n, misc::dbl_pair shapes[], size_t ids[],
	size_t rotated[], bool build_rcg)
	: acg_base_(n, shapes, ids), rotated_(n, 0), mapped_(false)
{
	if (rotated != 0)
		for (size_t i = 0; i < n; ++i)
		{
			rotated_[i] = rotated[i];
			if (rotated[i] != 0)
				acg_base_.node_rotate(i);
		}

	if (build_rcg)
		acg::rcg_build(acg_base_);
}

rcg_fp_t::~rcg_fp_t()
{
}

void rcg_fp_t::swap(rcg_fp_t &r)
{
	acg_base_.swap(r.acg_base_);
	rotated_.swap(r.rotated_);
	std::swap(shape_, r.shape_);
	std::swap(mapped_, r.mapped_);
}

perturb_ptr rcg_fp_t::create_perturbation(double rotateP, double exchangeP)
{
	return perturb_ptr(new rcg_perturb_t(this, rotateP, exchangeP));
}

void rcg_fp_t::do_packing()
{
	if (!mapped_)
	{
		shape_ = acg::packing_longest_path(acg_base_);

		mapped_ = true;
	}
}

void rcg_fp_t::compute_bb()
{
	do_packing();

	acg::longest_path_bb(acg_base_, shape_);
}

size_t rcg_fp_t::get_num_modules()
{
	return acg_base_.num_nodes();
}

misc::dbl_pair rcg_fp_t::get_shape()
{
	do_packing();

	return shape_;
}

misc::dbl_pair rcg_fp_t::get_centers(size_t ids[], misc::dbl_pair centers[])
{
	do_packing();

	if (ids == 0)
		for (size_t i = 0; i < acg_base_.num_nodes(); ++i)
			centers[acg_base_[i]->m_id] = acg_base_[i]->center(acg::POS_IN);
	else
		for (size_t i = 0; i < acg_base_.num_nodes(); ++i)
		{
			ids[i] = acg_base_[i]->m_id;
			centers[i] = acg_base_[i]->center(acg::POS_IN);
		}

	return shape_;
}

misc::dbl_pair rcg_fp_t::get_physical_info(
	size_t ids[], misc::dbl_pair centers[],
	misc::dbl_pair lls[], misc::dbl_pair urs[],
	size_t rotated[], misc::dbl_pair rotated_shapes[])
{
	compute_bb();

	for (size_t i = 0; i < acg_base_.num_nodes(); ++i)
	{
		size_t index = (ids == 0)? acg_base_[i]->m_id: i;

		acg::node_t *node = acg_base_[i];
		if (centers != 0)
			centers[index] = node->center(acg::POS_IN);
		if (lls != 0)
			lls[index] = node->bb(acg::POS_IN);
		if (urs != 0)
			urs[index] = node->bb(acg::POS_OUT);
		if (rotated_shapes != 0)
			rotated_shapes[index] = node->width();

		if (ids != 0)
			ids[index] = acg_base_[i]->m_id;

		if (rotated != 0)
			rotated[index] = rotated_[i];
	}

	return shape_;
}

bool rcg_fp_t::dump(const std::string &name, misc::cstr_t module_names[])
{
	compute_bb();

	FILE *fp = fopen(name.c_str(), "w");
	if (fp == 0)
		return false;

	fprintf(fp, "<?xml version=\"1.0\"?>\n");

	fprintf(fp,
		"<design x='%e' y='%e' width='%e' height='%e'>\n",
		0.0, 0.0, shape_.first, shape_.second);

	fprintf(fp, "  <aliases>\n");
	for (size_t i = 0; i < acg_base_.num_nodes(); ++i)
		if (module_names != 0)
			fprintf(fp,
				"    <alias module='%p' name='%s' />\n",
				acg_base_[i], module_names[acg_base_[i]->m_id]);
		else
			fprintf(fp,
				"    <alias module='%p' name='%u' />\n",
				acg_base_[i], acg_base_[i]->m_id);
	fprintf(fp, "  </aliases>\n");

	acg_base_.xml_write(fp);

	fprintf(fp, "</design>\n");

	fclose(fp);

	return true;
}

rcg_perturb_t::rcg_perturb_t(rcg_fp_t *fp, double rotateP, double exchangeP)
	: fp_(fp),
	rotateP_(rotateP), exchangeP_(exchangeP),
	choice_(-1), n1_(0), n2_(0)
{
}

void rcg_perturb_t::perturbation(double)
{
	const size_t n = fp_->acg_base_.num_nodes();

	assert(n != 0);

	double P = double(rand())/RAND_MAX;

	fp_->mapped_ = false;

	if ((P < rotateP_) || (n == 1))
	{
		choice_ = 0;
		
		n1_ = rand()%n;

		fp_->acg_base_.node_rotate(n1_);
		fp_->rotated_[n1_] = 1-fp_->rotated_[n1_];
	}
	else if (P < rotateP_+exchangeP_)
	{
		choice_ = 1;
	
		n1_ = rand()%fp_->acg_base_.num_nodes();
	
		do
		{
			n2_ = (int)rand()%n;
		}
		while (n2_ == n1_);

		fp_->acg_base_.node_swap(n1_, n2_);
		std::swap(fp_->rotated_[n1_], fp_->rotated_[n2_]);
	}
	else
	{
		choice_ = 2;

		n1_ = rand()%(n-1);

		acg::rcg_swap(fp_->acg_base_, n1_);
		std::swap(fp_->rotated_[n1_], fp_->rotated_[n1_+1]);
	}
}

void rcg_perturb_t::reject()
{
	fp_->mapped_ = false;

	switch (choice_)
	{
	case 0:
		fp_->acg_base_.node_rotate(n1_);
		fp_->rotated_[n1_] = 1-fp_->rotated_[n1_];
		break;
	case 1:
		fp_->acg_base_.node_swap(n1_, n2_);
		std::swap(fp_->rotated_[n1_], fp_->rotated_[n2_]);
		break;
	case 2:
		acg::rcg_swap(fp_->acg_base_, n1_);
		std::swap(fp_->rotated_[n1_], fp_->rotated_[n1_+1]);
		break;
	default:
		assert(false);
	}
}

void rcg_perturb_t::record_best()
{
	best_ = fp_->acg_base_;
	best_rotated_ = fp_->rotated_;
}

void rcg_perturb_t::swap_best()
{
	best_.swap(fp_->acg_base_);
	best_rotated_.swap(fp_->rotated_);
	fp_->mapped_ = false;
}

} // namespace fplan
