/*
 * Copyright (c) 2004-2005 Northwestern University.
 *                         NuCAD Group.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * geometry.h - 2D geometry computations.
 *
 * Authors: Jia Wang, jwa112@ece.northwestern.edu
 *
 */

#ifndef GEOMETRY_H
#define GEOMETRY_H

namespace misc
{

// we need a double pair to represent 2D objects
// although std::pair works, we write our own so operators
// could be put in this namespace

struct dbl_pair
{
	double first, second;

	dbl_pair(double f = 0, double s = 0) : first(f), second(s) {}

	template <class T, class U>
	dbl_pair(const std::pair<T, U> &p) : first(p.first), second(p.second) {}

	// operators
	dbl_pair &operator += (const dbl_pair &r)
		{first += r.first; second += r.second; return *this;}

	dbl_pair &operator -= (const dbl_pair &r)
		{first -= r.first; second -= r.second; return *this;}

	dbl_pair &operator *= (const dbl_pair &r)
		{first *= r.first; second *= r.second; return *this;}

	dbl_pair &operator /= (const dbl_pair &r)
		{first /= r.first; second /= r.second; return *this;}

	dbl_pair &operator += (double r)
		{first += r; second += r; return *this;}

	dbl_pair &operator -= (double r)
		{first -= r; second -= r; return *this;}

	dbl_pair &operator *= (double r)
		{first *= r; second *= r; return *this;}

	dbl_pair &operator /= (double r)
		{first /= r; second /= r; return *this;}

}; // struct dbl_pair

inline dbl_pair operator + (dbl_pair l, const dbl_pair &r) {l += r; return l;}
inline dbl_pair operator - (dbl_pair l, const dbl_pair &r) {l -= r; return l;}
inline dbl_pair operator * (dbl_pair l, const dbl_pair &r) {l *= r; return l;}
inline dbl_pair operator / (dbl_pair l, const dbl_pair &r) {l /= r; return l;}
inline dbl_pair operator + (dbl_pair l, double r) {l += r; return l;}
inline dbl_pair operator - (dbl_pair l, double r) {l -= r; return l;}
inline dbl_pair operator * (dbl_pair l, double r) {l *= r; return l;}
inline dbl_pair operator / (dbl_pair l, double r) {l /= r; return l;}
inline dbl_pair operator + (double l, const dbl_pair &r) {dbl_pair ll(l, l); ll += r; return ll;}
inline dbl_pair operator - (double l, const dbl_pair &r) {dbl_pair ll(l, l); ll -= r; return ll;}
inline dbl_pair operator * (double l, const dbl_pair &r) {dbl_pair ll(l, l); ll *= r; return ll;}
inline dbl_pair operator / (double l, const dbl_pair &r) {dbl_pair ll(l, l); ll /= r; return ll;}

inline dbl_pair max(const dbl_pair &l, const dbl_pair &r)
{
	return dbl_pair(
		(l.first < r.first)? r.first: l.first,
		(l.second < r.second)? r.second: l.second);
}

inline dbl_pair min(const dbl_pair &l, const dbl_pair &r)
{
	return dbl_pair(
		(l.first > r.first)? r.first: l.first,
		(l.second > r.second)? r.second: l.second);
}

} // namespace misc

#endif // GEOMETRY_H
